/***************************************************************************
                          phonebookcontrol.cpp  -  description
                             -------------------
    begin                : Thu Jan 18 2001
    copyright            : (C) 2001 by Matthias Welwarsky
    email                : matze@stud.fbi.fh-darmstadt.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <kdebug.h>
#include <klocale.h>
#include <klistview.h>

//#include <qlistview.h>
#include <qtabwidget.h>
#include <qlineedit.h>
#include <qprogressdialog.h>
#include <qcheckbox.h>

#include "phonectrlcenter.h"
#include "messagedispatcher.h"
#include "phonebookcontrol.h"
#include "phonebookitem.h"
#include "phonebookeditdlg.h"

const char* PhoneBookControl::bookDenominator[] = {
	"FD", "SM", "ME", "DC", "LD", "MC", "RC", "ON"
};
const char* PhoneBookControl::bookViewTitle[] =
	{ "Fixed Dials", "SIM", "Phone", "Last Dialled",
	  "Last Dialled", "Missed", "Received", "Own Numbers" };

PhoneBookControl::PhoneBookControl(QWidget *parent, const char *name )
	: PhoneBookControlDlg(parent,name), updateListIterator(updateList), indexBounds(8)
{

	PhoneCtrlCenter* pcCenter = dynamic_cast<PhoneCtrlCenter*>(parentWidget());
	if (!pcCenter) {
		kdDebug() << "cannot resolve control center\n";
		return;
	}
	messageDispatcher = pcCenter->getDispatcher();

	availableBooks = 0;

	connect(syncButton, SIGNAL(clicked()), this, SLOT(synchronizeBooks()));
	connect(buttonNewEntry, SIGNAL(clicked()), this, SLOT(newEntry()));
	connect(buttonDelete, SIGNAL(clicked()), this, SLOT(deleteEntry()));
}

PhoneBookControl::~PhoneBookControl(){
}

/** starts the phonebook control, reimplemented from PhoneControl */
void PhoneBookControl::startControl() {

	messageDispatcher->acquireDispatcher(this, SLOT(deliverMessage(QString)));
	
	if (!availableBooks)
		phoneBookView->removePage(phoneBookView->currentPage());

	selectedBook = FD;
	controlState = SelectingPhonebook;
	messageDispatcher->sendCommand(QString("AT+CPBS=\"")+selectedPhonebook()+QString("\""), this);
}

/** stops the phonebook control, reimplemented from PhoneControl */
void PhoneBookControl::stopControl() {

	messageDispatcher->releaseDispatcher(this);
	controlState = Idle;
}

/** reimplemented from PhoneControl */
void PhoneBookControl::deliverMessage(QString message) {
	int pos;
	int len;

	// drop unexpected stuff
	if (controlState == Idle) {
		//kdDebug() << "Ignoring spurious message\n";
		return;
	}

	if (message.find(QRegExp("ERROR")) != -1) {
		switch (controlState) {
		case SelectingPhonebook:
			// phonebook is not available, switch to the next book;
			kdDebug() << "book " << selectedPhonebook() << " is not available, switching to next\n";
			if (selectedBook < ON) {
				selectedBook += SM;
				messageDispatcher->sendCommand(QString("AT+CPBS=\"")+bookDenominator[selectedBook]+QString("\""), this);
			}
			break;

		default:
			// just ignore the error
			break;
		}
		return;
	}

	if (message.find(QRegExp("\nOK")) != -1) {
		switch (controlState) {
		case SelectingPhonebook: {
			// phonebook selected, add a tab for it
			kdDebug() << "Phonebook " << bookDenominator[selectedBook] << " selected OK\n";

			if (bookViewList.find(bookDenominator[selectedBook])) {
				kdDebug() << "already have view for this book\n";
				controlState = QueryingPhonebook;
				messageDispatcher->sendCommand("AT+CPBR=?", this);
				break;
			}

			// construct a view for the book
			KListView* bookView = new KListView();
			
			bookView->addColumn("");
			bookView->addColumn("Name");
			bookView->addColumn("Phone Number");
			bookView->setSorting(1);
			bookView->setShowSortIndicator(true);
			bookView->setAllColumnsShowFocus(true);
			bookViewList.insert(bookDenominator[selectedBook], bookView);

			// only certain books are writeable and thus editable
			if (selectedBook == SM || selectedBook == ME) {
				bookView->setItemsRenameable(true);
				bookView->setRenameable(0, false);
				bookView->setRenameable(1, true);
				bookView->setRenameable(2, true);
				
				connect(bookView, SIGNAL(itemRenamed(QListViewItem*, const QString&, int)),
						this, SLOT(itemRenamed(QListViewItem*, const QString&, int)));						
			}
			
			// map the widget to the phonebook
			bookViewMap[bookView] = selectedBook;

			phoneBookView->addTab(bookView, bookViewTitle[selectedBook]);
			phoneBookView->showPage(bookView);

			availableBooks++;

			controlState = QueryingPhonebook;
			messageDispatcher->sendCommand("AT+CPBR=?", this);
			break;
          }

		case QueryingPhonebook:
			// move to ReadingPhonebook
			controlState = ReadingPhonebook;
			break;

		case ReadingPhonebook:
			kdDebug() << "ReadingPhonebook finished, selecting next one\n";
			if (selectedBook < ON) {
				selectedBook += 1;
				controlState = SelectingPhonebook;
				messageDispatcher->sendCommand(QString("AT+CPBS=\"")+bookDenominator[selectedBook]+QString("\""), this);
				break;
			}
			kdDebug() << "gone through all books\n";
			controlState = Idle;
			break;

		case Synchronizing:
			++updateListIterator;
			++progress;
			if (updateListIterator.current() == NULL) {
				controlState = Idle;
				kdDebug() << "Sync finished\n";
				delete progressDialog;
				break;
			}
			kdDebug() << **updateListIterator << endl;
			progressDialog->setProgress(progress);
			messageDispatcher->sendCommand(**updateListIterator, this);
			break;

		default:
			// just ignore
			break;
		}
		return;
	}

	if (message.find(QRegExp("\\+CPBR:")) != -1) {
		// find a valid range of entries in the phone book.
		if ((pos = QRegExp("\\([0-9\\-]+\\)").match(message, 0, &len)) != -1) {
			pos = QRegExp("\\d+").match(message, pos, &len);
			QString startIndex = message.mid(pos, len);

			pos = QRegExp("\\d+").match(message, pos+len, &len);
			QString endIndex = message.mid(pos, len);

			// store the values for later use
			indexBounds[(int)selectedBook] = BookIndexBound(startIndex.toInt(), endIndex.toInt());
			
			// got the range, now query the book
			messageDispatcher->sendCommand("AT+CPBR="+startIndex+","+endIndex, this);
		} else {
			// not a range query, add an entry to the view.
			pos = QRegExp("\\d+").match(message, 0, &len);
			int index = message.mid(pos, len).toInt();

			pos = QRegExp("\"[\\+0-9]+\"").match(message, pos+len, &len);
			QString number = message.mid(pos+1, len-2);

			pos = QRegExp("\\d+").match(message, pos+len, &len);
			int type = message.mid(pos, len).toInt();

			pos = QRegExp("\".*\"").match(message, pos+len, &len);
			QString name = PhonebookItem::gsmUnescape(message.mid(pos+1, len-2));

			// avoid double entries
			QListView* bookView = bookViewList[bookDenominator[selectedBook]];
			PhonebookItem* item = static_cast<PhonebookItem*>(bookView->firstChild());

			while (item != NULL) {
				if (item->name() == name && item->number() == number)
					break;
				item = static_cast<PhonebookItem*>(item->nextSibling());
			}
			if (!item)
				new PhonebookItem(bookView, name, number, index, type);
		}
	}
}

/** this slot it called when someone activates a phonebook item */
void PhoneBookControl::editPhonebookItem(QListViewItem* item){
	PhonebookItem *entry = dynamic_cast<PhonebookItem*>(item);
	if (!entry) {
		kdDebug() << "Selected Entry is not a PhonebookItem\n";
		return;
	}

	PhonebookEditDialog* entryEditor = new PhonebookEditDialog(this, "itemEditor", true);
	entryEditor->nameEdit->setText(entry->name());
	entryEditor->numberEdit->setText(entry->number());

	if (entryEditor->exec()) {
		entry->setName(entryEditor->nameEdit->text());
		entry->setNumber(entryEditor->numberEdit->text());
		if (entryEditor->checkboxDeleteEntry->isChecked())
			entry->setState(PhonebookItem::DELETED);
	}
}

/** called when the "Synchronize..." button is clicked.
  * scans all local entries and writes back those that have changed.
  */
void PhoneBookControl::synchronizeBooks() {
	PhoneBook bookIndex;

	updateList.clear();
	
	for (bookIndex = SM; bookIndex < DC; bookIndex +=1) {
		QListView* bookView = bookViewList[bookDenominator[bookIndex]];
		QListViewItemIterator it(bookView);

		QString* command = new QString;
		*command = "AT+CPBS=\"" + QString(bookDenominator[bookIndex]) + "\"";
		updateList.append(command);
		
		for (; it.current(); ++it) {
			PhonebookItem* item = static_cast<PhonebookItem*>(it.current());
			
			if (item->changed()) {
				QString* entry = new QString();
	
				// check if the item is to be deleted
				// append the erase command to the list and delete the item from the view			
				if (item->state() == PhonebookItem::DELETED) {
					entry->sprintf("AT+CPBW=%d", item->index());
					updateList.append(entry);
					delete item;
					continue;
				}
					
				// if the number is prefixed with "+", the type needs to be
				// 154. Else, it needs to be 129;
				int itemType = 129;
				if (item->number()[0] == '+')
					itemType = 145;

				entry->sprintf("AT+CPBW=%d,\"%s\",%d,\"%s\"",
					item->index(),
					(const char*)PhonebookItem::gsmEscape(item->number()),
					itemType,
					(const char*)PhonebookItem::gsmEscape(item->name()));

				updateList.append(entry);
				item->setState(PhonebookItem::CLEAN);
			}
		}
	}
	sendUpdates();		
}

/** transmits a list of updates to the mobile phone. */
void PhoneBookControl::sendUpdates(){
	// sanity check
	if (updateList.isEmpty())
		return;

	controlState = Synchronizing;
	updateListIterator.toFirst();

	kdDebug() << **updateListIterator << endl;

	progress = 1;
	progressDialog = new QProgressDialog(this);
	progressDialog->setLabelText(i18n("Synchronizing..."));
	progressDialog->setTotalSteps(updateList.count());
	progressDialog->setProgress(progress);

	messageDispatcher->sendCommand(**updateListIterator, this);
}

/** conveniece function. returns QString(bookDenominator(selectedBook)). */
QString PhoneBookControl::selectedPhonebook(){
	return QString(bookDenominator[selectedBook]);
}

/** This slot is called when the "New Entry" button is
called. It then constructs an empty PhonebookItem
and opens the an editor window. */
void PhoneBookControl::newEntry(){
	// find out the book to add the entry to
	KListView* bookView = (KListView*)phoneBookView->currentPage();
	PhonebookItem *entry = new PhonebookItem(bookView);
	
	int upperBound = indexBounds[bookViewMap[bookView]].upperBound;
	
	// find a free position in the phonebook. don't care about sorting,
	// the ListView does this.
	// FIXME: how many entries may the book hold? this can be queried
	// from the phone, DO THIS!
	QArray<PhonebookItem*> itemArray(upperBound);
	QListViewItemIterator it(bookView);
	for (; it.current(); ++it) {
		PhonebookItem* item = static_cast<PhonebookItem*>(it.current());
		kdDebug() << "found entry " << item->index() << endl;
		itemArray[item->index()] = item;
	}
	
	int nextFreePosition;
	for (nextFreePosition = 0; nextFreePosition < upperBound; nextFreePosition++) {
		if (itemArray[nextFreePosition] == NULL)
			break;
	}
	if (nextFreePosition == upperBound) {
		// FIXME: display a message box
		return;
	}
	
	PhonebookEditDialog* entryEditor = new PhonebookEditDialog(this, "itemEditor", true);
	
	// we reuse the editor but we do not want the user do "Delete" an entry
	// that's not really in existance
	entryEditor->checkboxDeleteEntry->setEnabled(false);

	if (entryEditor->exec()) {
		entry->setName(entryEditor->nameEdit->text());
		entry->setNumber(entryEditor->numberEdit->text());
		entry->setIndex(nextFreePosition);
	}
}
/** connected to the KlistView, called when an item
is renamed. */
void PhoneBookControl::itemRenamed(QListViewItem* item, const QString& newContent, int col){
	PhonebookItem* entry = static_cast<PhonebookItem*>(item);

	switch (col) {
	case 1:
		entry->setName(newContent);
		break;
		
	case 2:
		entry->setNumber(newContent);
		break;
	}
}

/** called when the "Delete" button is clicked. */
void PhoneBookControl::deleteEntry(){
	KListView* bookView = (KListView*)phoneBookView->currentPage();
	PhonebookItem* entry = static_cast<PhonebookItem*>(bookView->selectedItem());
	if (entry == NULL)
		return;
	
	if (entry->state() == PhonebookItem::DELETED)
		entry->setState(PhonebookItem::CLEAN);
	else
		entry->setState(PhonebookItem::DELETED);
}
